# -*- coding: utf-8 -*-

def hamming(seq1, seq2, normalized=False):
	"""Compute the Hamming distance between the two sequences `seq1` and `seq2`.
	The Hamming distance is the number of differing items in two ordered
	sequences of the same length. If the sequences submitted do not have the
	same length, an error will be raised.
	
	If `normalized` evaluates to `False`, the return value will be an integer
	between 0 and the length of the sequences provided, edge values included;
	otherwise, it will be a float between 0 and 1 included, where 0 means
	equal, and 1 totally different. Normalized hamming distance is computed as:
	
		0.0                         if len(seq1) == 0
		hamming_dist / len(seq1)    otherwise
	"""
	L = len(seq1)
	if L != len(seq2):
		raise ValueError("expected two strings of the same length")
	if L == 0:
		return 0.0 if normalized else 0  # equal
	dist = sum(c1 != c2 for c1, c2 in zip(seq1, seq2))
	if normalized:
		return dist / float(L)
	return dist

def jaccard(seq1, seq2):
	"""Compute the Jaccard distance between the two sequences `seq1` and `seq2`.
	They should contain hashable items.
	
	The return value is a float between 0 and 1, where 0 means equal, and 1 totally different.
	"""
	set1, set2 = set(seq1), set(seq2)
	return 1 - len(set1 & set2) / float(len(set1 | set2))


def sorensen(seq1, seq2):
	"""Compute the Sorensen distance between the two sequences `seq1` and `seq2`.
	They should contain hashable items.
	
	The return value is a float between 0 and 1, where 0 means equal, and 1 totally different.
	"""
	set1, set2 = set(seq1), set(seq2)
	return 1 - (2 * len(set1 & set2) / float(len(set1) + len(set2)))



from ._pyimports import levenshtein, fast_comp

def ilevenshtein(seq1, seqs, max_dist=-1):
	"""Compute the Levenshtein distance between the sequence `seq1` and the series
	of	sequences `seqs`.
	
		`seq1`: the reference sequence
		`seqs`: a series of sequences (can be a generator)
		`max_dist`: if provided and > 0, only the sequences which distance from
		the reference sequence is lower or equal to this value will be returned.
	
	The return value is a series of pairs (distance, sequence).
	
	The sequence objects in `seqs` are expected to be of the same kind than
	the reference sequence in the C implementation; the same holds true for
	`ifast_comp`.
	"""
	for seq2 in seqs:
		dist = levenshtein(seq1, seq2, max_dist=max_dist)
		if dist != -1:
			yield dist, seq2


def ifast_comp(seq1, seqs, transpositions=False):
	"""Return an iterator over all the sequences in `seqs` which distance from
	`seq1` is lower or equal to 2. The sequences which distance from the
	reference sequence is higher than that are dropped.
	
		`seq1`: the reference sequence.
		`seqs`: a series of sequences (can be a generator)
		`transpositions` has the same sense than in `fast_comp`.
	
	The return value is a series of pairs (distance, sequence).
	
	You might want to call `sorted()` on the iterator to get the results in a
	significant order:
	
		>>> g = ifast_comp("foo", ["fo", "bar", "foob", "foo", "foobaz"])
		>>> sorted(g)
		[(0, 'foo'), (1, 'fo'), (1, 'foob')]
	"""
	for seq2 in seqs:
		dist = fast_comp(seq1, seq2, transpositions)
		if dist != -1:
			yield dist, seq2



import os, sys
from array import array
try:
	from distance import cdistance
except ImportError:
	cdistance = None
from distance import _pyimports as pydistance


if sys.version_info.major < 3:
	t_unicode = unicode
	t_bytes = lambda s: s
else:
	t_unicode = lambda s: s
	t_bytes = lambda s: s.encode()

all_types = [
	("unicode", t_unicode),
	("bytes", t_bytes),
	("list", list),
	("tuple", tuple),
]


def hamming(func, t, **kwargs):

	# types; only for c
	if kwargs["lang"] == "C":
		try:
			func(1, t("foo"))
		except ValueError:
			pass
		try:
			func(t("foo"), 1)
		except ValueError:
			pass

	# empty string
	assert func(t(""), t("")) == 0
	
	# common
	assert func(t("abc"), t("abc")) == 0
	assert func(t("abc"), t("abd")) == 1
	
	# wrong length
	try:
		func(t("foo"), t("foobar"))
	except ValueError:
		pass

	try:
		func(t(""), t("foo"))
	except ValueError:
		pass
	
	# normalization
	assert func(t(""), t(""), normalized=True) == 0.0
	assert func(t("abc"), t("abc"), normalized=True) == 0.0
	assert func(t("ab"), t("ac"), normalized=True) == 0.5
	assert func(t("abc"), t("def"), normalized=True) == 1.0


def fast_comp(func, t, **kwargs):

	# types; only for c
	if kwargs["lang"] == "C":
		try:
			func(1, t("foo"))
		except ValueError:
			pass
		try:
			func(t("foo"), 1)
		except ValueError:
			pass

	# empty strings
	assert func(t(""), t("")) == 0
	assert func(t(""), t("a")) == func(t("a"), t("")) == 1

	# edit ops
	assert func(t("aa"), t("aa")) == 0
	assert func(t("ab"), t("aa")) == 1
	assert func(t("ab"), t("a")) == 1
	assert func(t("ab"), t("abc")) == 1
	
	# dist limit
	assert func(t("a"), t("bcd")) == func(t("bcd"), t("a")) == -1
	
	# transpositions
	assert func(t("abc"), t("bac"), transpositions=True) == \
		func(t("bac"), t("abc"), transpositions=True) == 1
	


def levenshtein(func, t, **kwargs):

	# types; only for c
	if kwargs["lang"] == "C":
		try:
			func(1, t("foo"))
		except ValueError:
			pass
		try:
			func(t("foo"), 1)
		except ValueError:
			pass

	# empty strings
	assert func(t(""), t("")) == 0
	assert func(t(""), t("abcd")) == func(t("abcd"), t("")) == 4
	
	# edit ops
	assert func(t("aa"), t("aa")) == 0
	assert func(t("ab"), t("aa")) == 1
	assert func(t("ab"), t("a")) == 1
	assert func(t("ab"), t("abc")) == 1
	
	# dist limit
	assert func(t("a"), t("b"), max_dist=0) == -1
	assert func(t("a"), t("b"), max_dist=1) == 1
	assert func(t("foo"), t("bar"), max_dist=-1) == 3


def nlevenshtein(func, t, **kwargs):

	# types; only for c
	if kwargs["lang"] == "C":
		try:
			func(1, t("foo"))
		except ValueError:
			pass
		try:
			func(t("foo"), 1)
		except ValueError:
			pass

	# empty strings
	assert func(t(""), t(""), 1) == func(t(""), t(""), 2) == 0.0
	assert func(t(""), t("foo"), 1) == func(t("foo"), t(""), 1) == \
		func(t(""), t("foo"), 2) == func(t("foo"), t(""), 2) == 1.0

	assert func(t("aa"), t("aa"), 1) == func(t("aa"), t("aa"), 2) == 0.0
	assert func(t("ab"), t("aa"), 1) == func(t("ab"), t("aa"), 2) == 0.5
	assert func(t("ab"), t("a"), 1) == func(t("ab"), t("a"), 2) == 0.5
	assert func(t("ab"), t("abc"), 1) == func(t("ab"), t("abc"), 2) == 0.3333333333333333

	# multiple alignments
	assert func(t("abc"), t("adb"), 1) == 0.6666666666666666
	assert func(t("abc"), t("adb"), 2) == 0.5


def lcsubstrings(func, t, **kwargs):

	# types; only for c
	if kwargs["lang"] == "C":
		try:
			func(1, t("foo"))
		except ValueError:
			pass
		try:
			func(t("foo"), 1)
		except ValueError:
			pass

	# empty strings
	try:
		assert func(t(""), t(""), False) == set()
	except TypeError:
		if t is not list: raise
	assert func(t(""), t(""), True) == (0, ())
	try:
		assert func(t(""), t("foo"), False) == func(t("foo"), t(""), False) == set()
	except TypeError:
		if t is not list: raise
	assert func(t(""), t("foo"), True) == func(t("foo"), t(""), True) == (0, ())
	
	# common
	try:
		assert func(t("abcd"), t("cdba"), False) == {t('cd')}
	except TypeError:
		if t is not list: raise
	assert func(t("abcd"), t("cdba"), True) == (2, ((2, 0),))
	
	# reverse
	try:
		assert func(t("abcdef"), t("cdba"), False) == func(t("cdba"), t("abcdef"), False)
	except TypeError:
		if t is not list: raise
	assert func(t("abcdef"), t("cdba"), True) == func(t("cdba"), t("abcdef"), True)


def itors_common(func, t, **kwargs):

	if kwargs["lang"] == "C":
		# types check; only need to do it for C impl to avoid an eventual segfaults.
		try: func(1, t("foo"))
		except ValueError: pass
	
		itor = func(t("foo"), [t("foo"), 3333])
		next(itor)
		try: next(itor)
		except ValueError: pass

	# values drop
	itor = func(t("aa"), [t("aa"), t("abcd"), t("ba")])
	assert next(itor) == (0, t("aa"))
	assert next(itor) == (1, t("ba"))


def ilevenshtein(func, t, **kwargs):
	itors_common(lambda a, b: func(a, b, max_dist=2), t, **kwargs)
	

def ifast_comp(func, t, **kwargs):
	itors_common(func, t, **kwargs)
	#transpositions
	g = func(t("abc"), [t("bac")], transpositions=False)
	assert next(g) == (2, t('bac'))
	g = func(t("abc"), [t("bac")], transpositions=True)
	assert next(g) == (1, t("bac"))
	

write = lambda s: sys.stderr.write(s + '\n')

tests = ["hamming", "fast_comp", "levenshtein", "lcsubstrings", "nlevenshtein", "ilevenshtein", "ifast_comp"]


def run_test(name):
	if cdistance:
		cfunc = getattr(cdistance, name)
		run_lang_test(name, cfunc, "C")
		write("")
	pyfunc = getattr(pydistance, name)
	run_lang_test(name, pyfunc, "py")
	if cdistance is None:
		write("skipped C tests")
	write("")


def run_lang_test(name, func, lang):
	print("%s (%s)..." % (name, lang))
	for tname, typ in all_types:
		write("type: %s" % tname)
		globals()[name](func, typ, lang=lang)

if __name__ == "__main__":
	args = sys.argv[1:]
	if not args:
		for test in tests:
			run_test(test)
		sys.exit()
	for name in args:
		if name in tests:
			run_test(name)
		else:
			write("no such test: %s" % name)
			sys.exit(1)



# -*- coding: utf-8 -*-

from array import array


def levenshtein(seq1, seq2, normalized=False, max_dist=-1):
	"""Compute the absolute Levenshtein distance between the two sequences
	`seq1` and `seq2`.
	
	The Levenshtein distance is the minimum number of edit operations necessary
	for transforming one sequence into the other. The edit operations allowed are:
	
		* deletion:     ABC -> BC, AC, AB
		* insertion:    ABC -> ABCD, EABC, AEBC..
		* substitution: ABC -> ABE, ADC, FBC..
	
	The `max_dist` parameter controls at which moment we should stop computing the
	distance between the provided sequences. If it is a negative integer, the
	distance will be computed until the sequences are exhausted; otherwise, the
	computation will stop at the moment the calculated distance is higher than
	`max_dist`, and then return -1. For example:
	
		>>> levenshtein("abc", "abcd", max_dist=1)  # dist = 1
		1
		>>> levenshtein("abc", "abcde", max_dist=1) # dist = 2
		-1
	
	This can be a time saver if you're not interested in the exact distance, but
	only need to check if the distance between the given sequences is below a
	given threshold.
	
	The `normalized` parameter is here for backward compatibility; providing
	it will result in a call to `nlevenshtein`, which should be used directly
	instead. 
	"""
	if normalized:
		return nlevenshtein(seq1, seq2, method=1)
		
	if seq1 == seq2:
		return 0
	
	len1, len2 = len(seq1), len(seq2)
	if max_dist >= 0 and abs(len1 - len2) > max_dist:
		return -1
	if len1 == 0:
		return len2
	if len2 == 0:
		return len1
	if len1 < len2:
		len1, len2 = len2, len1
		seq1, seq2 = seq2, seq1
	
	column = array('L', range(len2 + 1))
	
	for x in range(1, len1 + 1):
		column[0] = x
		last = x - 1
		for y in range(1, len2 + 1):
			old = column[y]
			cost = int(seq1[x - 1] != seq2[y - 1])
			column[y] = min(column[y] + 1, column[y - 1] + 1, last + cost)
			last = old
		if max_dist >= 0 and min(column) > max_dist:
			return -1
	
	if max_dist >= 0 and column[len2] > max_dist:
		# stay consistent, even if we have the exact distance
		return -1
	return column[len2]


def nlevenshtein(seq1, seq2, method=1):
	"""Compute the normalized Levenshtein distance between `seq1` and `seq2`.
	
	Two normalization methods are provided. For both of them, the normalized
	distance will be a float between 0 and 1, where 0 means equal and 1
	completely different. The computation obeys the following patterns:
	
		0.0                       if seq1 == seq2
		1.0                       if len(seq1) == 0 or len(seq2) == 0
		edit distance / factor    otherwise
	
	The `method` parameter specifies which normalization factor should be used.
	It can have the value 1 or 2, which correspond to the following:
	
		1: the length of the shortest alignment between the sequences
		   (that is, the length of the longest sequence)
		2: the length of the longest alignment between the sequences
	
	Which normalization factor should be chosen is a matter of taste. The first
	one is cheap to compute. The second one is more costly, but it accounts
	better than the first one for parallelisms of symbols between the sequences.
		
	For the rationale behind the use of the second method, see:
	Heeringa, "Measuring Dialect Pronunciation Differences using Levenshtein
	Distance", 2004, p. 130 sq, which is available online at:
	http://www.let.rug.nl/~heeringa/dialectology/thesis/thesis.pdf
	"""
	
	if seq1 == seq2:
		return 0.0
	len1, len2 = len(seq1), len(seq2)
	if len1 == 0 or len2 == 0:
		return 1.0
	if len1 < len2: # minimize the arrays size
		len1, len2 = len2, len1
		seq1, seq2 = seq2, seq1
	
	if method == 1:
		return levenshtein(seq1, seq2) / float(len1)
	if method != 2:
		raise ValueError("expected either 1 or 2 for `method` parameter")
	
	column = array('L', range(len2 + 1))
	length = array('L', range(len2 + 1))
	
	for x in range(1, len1 + 1):
	
		column[0] = length[0] = x
		last = llast = x - 1
		
		for y in range(1, len2 + 1):
		
			# dist
			old = column[y]
			ic = column[y - 1] + 1
			dc = column[y] + 1
			rc = last + (seq1[x - 1] != seq2[y - 1])
			column[y] = min(ic, dc, rc)
			last = old
			
			# length
			lold = length[y]
			lic = length[y - 1] + 1 if ic == column[y] else 0
			ldc = length[y] + 1 if dc == column[y] else 0
			lrc = llast + 1 if rc == column[y] else 0
			length[y] = max(ldc, lic, lrc)
			llast = lold
	
	return column[y] / float(length[y])



# -*- coding: utf-8 -*-

from array import array


def lcsubstrings(seq1, seq2, positions=False):
	"""Find the longest common substring(s) in the sequences `seq1` and `seq2`.
	
	If positions evaluates to `True` only their positions will be returned,
	together with their length, in a tuple:
	
		(length, [(start pos in seq1, start pos in seq2)..])
	
	Otherwise, the substrings themselves will be returned, in a set.
	
	Example:
	
		>>> lcsubstrings("sedentar", "dentist")
		{'dent'}
		>>> lcsubstrings("sedentar", "dentist", positions=True)
		(4, [(2, 0)])
	"""
	L1, L2 = len(seq1), len(seq2)
	ms = []
	mlen = last = 0
	if L1 < L2:
		seq1, seq2 = seq2, seq1
		L1, L2 = L2, L1
	
	column = array('L', range(L2))
	
	for i in range(L1):
		for j in range(L2):
			old = column[j]
			if seq1[i] == seq2[j]:
				if i == 0 or j == 0:
					column[j] = 1
				else:
					column[j] = last + 1
				if column[j] > mlen:
					mlen = column[j]
					ms = [(i, j)]
				elif column[j] == mlen:
					ms.append((i, j))
			else:
				column[j] = 0
			last = old
	
	if positions:
		return (mlen, tuple((i - mlen + 1, j - mlen + 1) for i, j in ms if ms))
	return set(seq1[i - mlen + 1:i + 1] for i, _ in ms if ms)



"Utilities for comparing sequences"

__all__ = ["hamming", "levenshtein", "nlevenshtein", "jaccard", "sorensen",
	"fast_comp", "lcsubstrings", "ilevenshtein", "ifast_comp"]

try:
	from .cdistance import *
except ImportError:
	from ._pyimports import *

from ._pyimports import jaccard, sorensen

def quick_levenshtein(str1, str2):
	return fast_comp(str1, str2, transpositions=False)

def iquick_levenshtein(str1, strs):
	return ifast_comp(str1, str2, transpositions=False)



# -*- coding: utf-8 -*- 

# Distance - Utilities for comparing sequences
# Copyright (C) 2013 Michaël Meyer

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import os, sys, ast, _ast, re
from distutils.core import setup, Extension

this_dir = os.path.dirname(os.path.abspath(__file__))
pkg_dir  = os.path.join(this_dir, "distance")
cpkg_dir  = os.path.join(this_dir, "cdistance")

ctypes = ["unicode", "byte", "array"]

cfunctions = {
	"levenshtein": ["levenshtein", "nlevenshtein"],
	"hamming": ["hamming"],
	"lcsubstrings": ["lcsubstrings"],
	"fastcomp": ["fastcomp"],
}

sequence_compare = """\
#define SEQUENCE_COMPARE(s1, i1, s2, i2) \\
(PyObject_RichCompareBool( \\
	PySequence_Fast_GET_ITEM((s1), (i1)), \\
	PySequence_Fast_GET_ITEM((s2), (i2)), \\
	Py_EQ) \\
)
"""

def make_c_doc():
	buff = []
	py_sources = [f for f in os.listdir(pkg_dir) if f.endswith('.py')]
	for file in py_sources:
		with open(os.path.join(pkg_dir, file)) as f:
			content = f.read()
		tree = ast.parse(content)
		for doc_string in parse_tree(tree, content):
			buff.append(doc_string)
	join_str = 2 * '\n'
	return join_str.join(buff) + '\n'


def parse_tree(tree, content):
	for node in ast.iter_child_nodes(tree):
		if not isinstance(node, _ast.FunctionDef):
			continue
		doc_string = ast.get_docstring(node)
		if not doc_string:
			continue
		func_def = re.findall("def\s%s\s*(.+?)\s*:" % node.name, content)
		assert func_def and len(func_def) == 1
		func_def = node.name + func_def[0] + 2 * '\\n\\\n'
		doc_string = doc_string.replace('\n', '\\n\\\n').replace('"', '\\"')
		doc_string = doc_string.replace('\n' + 8 * ' ', '\n' + 4 * ' ')
		doc_string = '#define %s_doc \\\n"%s%s"\n' % (node.name, func_def, doc_string)
		yield doc_string


def format_header():
	yield sequence_compare
	for cfile, cfuncs in cfunctions.items():
		for ctype in ctypes:
			if ctype == "array":
				yield("#define SEQUENCE_COMP SEQUENCE_COMPARE")
			yield('#define unicode %(type)s' % dict(type=ctype))
			for cfunc in cfuncs:
				yield("#define %(function)s %(tcode)s%(function)s" % dict(function=cfunc, tcode=ctype[0]))
			yield('#include "%(file)s.c"' % dict(file=cfile))
			yield("#undef unicode")
			for cfunc in cfuncs:
				yield("#undef %(function)s" % dict(function=cfunc))
			if ctype == "array":
				yield("#undef SEQUENCE_COMP")
			yield("")


def prepare():
	with open(os.path.join(cpkg_dir, "includes.h"), "w") as f:
		f.write(make_c_doc())
		f.write(4 * '\n')
		f.write('\n'.join(format_header()))


args = sys.argv[1:]
if "prepare" in args:
	prepare()
	sys.exit()

if "--with-c" in args:
	args.remove("--with-c")
	ext_modules = [Extension('distance.cdistance', sources=["cdistance/distance.c"])]
else:
	sys.stderr.write("notice: no C support available\n")
	ext_modules = []

with open(os.path.join(this_dir, "README.md")) as f:
    long_description = f.read()

setup (
    name = 'Distance',
    version = '0.1.3',
    description = 'Utilities for comparing sequences',
    long_description = long_description,
    author='Michaël Meyer',
    author_email='michaelnm.meyer@gmail.com',
    url='https://github.com/doukremt/distance',
    ext_modules = ext_modules,
    script_args = args,
    packages = ['distance'],
    classifiers=(
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
    )
)



# -*- coding: utf-8 -*-

def fast_comp(seq1, seq2, transpositions=False):
	"""Compute the distance between the two sequences `seq1` and `seq2` up to a
	maximum of 2 included, and return it. If the edit distance between the two
	sequences is higher than that, -1 is returned.
	
	If `transpositions` is `True`, transpositions will be taken into account for
	the computation of the distance. This can make a difference, e.g.:

		>>> fast_comp("abc", "bac", transpositions=False)
		2
		>>> fast_comp("abc", "bac", transpositions=True)
		1
	
	This is faster than `levenshtein` by an order of magnitude, but on the
	other hand is of limited use.

	The algorithm comes from `http://writingarchives.sakura.ne.jp/fastcomp`.
	I've added transpositions support to the original code.
	"""
	replace, insert, delete = "r", "i", "d"

	L1, L2  = len(seq1), len(seq2)
	if L1 < L2:
		L1, L2 = L2, L1
		seq1, seq2 = seq2, seq1

	ldiff = L1 - L2
	if ldiff == 0:
		models = (insert+delete, delete+insert, replace+replace)
	elif ldiff == 1:
		models = (delete+replace, replace+delete)
	elif ldiff == 2:
		models = (delete+delete,)
	else:
		return -1

	res = 3
	for model in models:
		i = j = c = 0
		while (i < L1) and (j < L2):
			if seq1[i] != seq2[j]:
				c = c+1
				if 2 < c:
					break
            
				if transpositions and ldiff != 2 \
            	and i < L1 - 1 and j < L2 - 1 \
            	and seq1[i+1] == seq2[j] and seq1[i] == seq2[j+1]:
					i, j = i+2, j+2
				else:
					cmd = model[c-1]
					if cmd == delete:
						i = i+1
					elif cmd == insert:
						j = j+1
					else:
						assert cmd == replace
						i,j = i+1, j+1
			else:
				i,j = i+1, j+1

		if 2 < c:
			continue
		elif i < L1:
			if L1-i <= model[c:].count(delete):
				c = c + (L1-i)
			else:
				continue
		elif j < L2:
			if L2-j <= model[c:].count(insert):
				c = c + (L2-j)
			else:
				continue

		if c < res:
			res = c

	if res == 3:
		res = -1
	return res
